/*file window.c
 * \brief GL4Dummies, exemple simple 3D avec GL4Dummies
 * \author Farès BELHADJ, amsi@up8.edu
 * \date February 03 2022
 */

/* inclusion des entêtes de fonctions de création et de gestion de
 * fenêtres système ouvrant un contexte favorable à GL4dummies. Cette
 * partie est dépendante de la bibliothèque SDL2 */
#include <GL4D/gl4duw_SDL2.h>
#include <GL4D/gl4dm.h>
#include <GL4D/gl4dg.h>

static void init(void);
static void draw(void);
static void quit(void);
static void resize(int width, int height);

/* dimensions de la fenêtre */
int _ww=800, _wh=600;

/* on créé une variable pour stocker l'identifiant du programme GPU */
GLuint _pId = 0;
/* on créé une variable pour stocker l'identifiant de la couleur */
GLuint _color = 0;
/* on créé une variable pour stocker l'identifiant de la géométrie d'un cube GL4D */
GLuint _cube = 0;
/* on créé une variable pour stocker l'identifiant de la géométrie d'un tore GL4D */
GLuint _torus = 0;
/* flag pour switcher d'une multi-vues à mono-vue (perspective) */
GLboolean _multi_vues = GL_FALSE;
/* un identifiant pour une texture */
GLuint _texId = 0;

/*!\brief créé la fenêtre, un screen 2D effacé en noir et lance une
 *  boucle infinie.*/
int main(int argc, char ** argv) {
  /* tentative de création d'une fenêtre pour GL4Dummies */
  if(!gl4duwCreateWindow(argc, argv, /* args du programme */
			 "GL4Dummies' Hello World", /* titre */
			 10, 10, _ww, _wh, /* x,y, largeur, heuteur */
			 GL4DW_SHOWN | GL4DW_RESIZABLE) /* état visible et redimensionnable */) {
    /* ici si échec de la création souvent lié à un problème d'absence
     * de contexte graphique ou d'impossibilité d'ouverture d'un
     * contexte OpenGL (au moins 3.2) */
    return 1;
  }
  /* appeler init pour initialiser des paramètres GL et GL4D */
  init();
  /* placer quit comme fonction à appeler au moment du exit */
  atexit(quit);
  /* placer draw comme fonction à appeler pour dessiner chaque frame */
  gl4duwDisplayFunc(draw);
  /* boucle infinie pour éviter que le programme ne s'arrête et ferme
   * la fenêtre immédiatement */
  /* placer resize comme fonction à appeler au redimensionnement de la fenêtre */
  gl4duwResizeFunc(resize);
  gl4duwMainLoop();
  return 0;
}

void resize(int width, int height) {
  GLfloat inv_ratio;
  _ww = width;
  _wh = height;
  /* déterminer la portion de la fenêtre dans laquelle on rend la vue */
  glViewport(0, 0, _ww, _wh);
  /* calculer l'inverse du ratio fenêtre (multiplicateur de la hauteur du plan near) */
  inv_ratio = _wh / (GLfloat)_ww;
  /* binder (mettre au premier plan, "en courante" ou "en active") la matrice proj */
  gl4duBindMatrix("proj");
  /* mettre la matrice identité (celle qui ne change rien) dans la matrice courante */
  gl4duLoadIdentityf();
  /* combiner la matrice courante avec une matrice de projection en
     perspective. Voir le support de cours pour les six paramètres :
     left, right, bottom, top, near, far */
  gl4duFrustumf(-1.0f, 1.0f, -1.0f * inv_ratio, 1.0f * inv_ratio, 1.0f, 1000.0f);
}

/* initialise des paramètres GL et GL4D */
void init(void) {
  GLuint image[] = { (0xFF << 24) | 0xFF, (0xFF << 24) | (0xFF << 8) | 0xFF, (0xFF << 24) | (0xFF << 8), -1 }    ;
  
  /* générer un identifiant de texture */
  glGenTextures(1, &_texId);
  assert(_texId);
  /* binder _texId comme texture bidimensionnelle */
  glBindTexture(GL_TEXTURE_2D, _texId);
  /* comportement de l'interpolation en minimisant puis maximisant la texture */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  /* spécification et transfert de la donnée */
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 2, 2, 0, GL_RGBA, GL_UNSIGNED_BYTE, image);
  /* dé-binder la texture bidimensionnelle */
  glBindTexture(GL_TEXTURE_2D, 0);

  /* set la couleur d'effacement OpenGL */
  glClearColor(0.0f, 0.0f, 0.5f, 1.0f);
  /* générer un cube en GL4D */
  _cube = gl4dgGenCubef();
  _torus = gl4dgGenTorusf(10, 5, 0.25);
  /* créer un programme GPU pour OpenGL (en GL4D) */
  _pId = gl4duCreateProgram("<vs>shaders/hello.vs", "<fs>shaders/hello.fs", NULL);
  _color = glGetUniformLocation(_pId, "color");
  /* créer dans GL4D une matrice qui s'appelle modview ; matrice de
     modélisation et vue qu'on retrouvera dans le vertex shader */
  gl4duGenMatrix(GL_FLOAT, "modview");
  /* créer dans GL4D une matrice qui s'appelle proj ; matrice de
     projection qu'on retrouvera dans le vertex shader */
  gl4duGenMatrix(GL_FLOAT, "proj");
  gl4duGenMatrix(GL_FLOAT, "view");
  resize(_ww, _wh);
  glEnable(GL_DEPTH_TEST);
}

void vue (int v){
  switch (v){
    case 0: 
      glViewport(0, 0, _ww / 2, _wh / 2);
      gl4duBindMatrix("proj");
      gl4duLoadIdentityf();
      gl4duOrthof(-5.0f, 5.0f, -5.0f *_wh / (GLfloat)_ww, 5.0f * _wh / (GLfloat) _wh, 0.0f, 1000.0f);
      gl4duBindMatrix("view");
      gl4duLoadIdentityf();
      gl4duLookAtf(0, 0, 5, 0, 0, 0, 0,1, 0);
      
      break;
    case 1: 
      glViewport(0, _wh/2, _ww / 2, _wh / 2);
      gl4duBindMatrix("proj");
      gl4duLoadIdentityf();
      gl4duOrthof(-5.0f, 5.0f, -5.0f *_wh / (GLfloat)_ww, 5.0f * _wh / (GLfloat) _wh, 0.0f, 1000.0f);
      gl4duBindMatrix("view");
      gl4duLoadIdentityf();
      gl4duLookAtf(5, 0, 0, 0, 0, 0, 0, 1, 0);
      break;
    case 2: 
      glViewport(_ww/2, _wh/2, _ww / 2, _wh / 2);
      gl4duBindMatrix("proj");
      gl4duLoadIdentityf();
      gl4duOrthof(-5.0f, 5.0f, -5.0f *_wh / (GLfloat)_ww, 5.0f * _wh / (GLfloat) _wh, 0.0f, 1000.0f);
      gl4duBindMatrix("view");
      gl4duLoadIdentityf();
      gl4duLookAtf(5, 0, 0, 0, 0, 0, 0, 0, -1);
      break;
   case 3: 
      glViewport(_ww/2, 0, _ww / 2, _wh / 2);
      gl4duBindMatrix("proj");
      gl4duLoadIdentityf();
      gl4duOrthof(-5.0f, 5.0f, -5.0f *_wh / (GLfloat)_ww, 5.0f * _wh / (GLfloat) _wh, 0.0f, 1000.0f);
      gl4duBindMatrix("view");
      gl4duLoadIdentityf();
      gl4duLookAtf(0, 2, 5, 0, 0, 0, 0, 1, 0);
      break;

   default: /* mono-vue */
     /* déterminer la portion de la fenêtre dans laquelle on rend la vue 0 */
     glViewport(0, 0, _ww, _wh);
     /* binder (mettre au premier plan, "en courante" ou "en active") la matrice proj */
     gl4duBindMatrix("proj");
     /* mettre la matrice identité (celle qui ne change rien) dans la matrice courante */
     gl4duLoadIdentityf();
     /* combiner la matrice courante avec une matrice de projection en
        perspective. Voir le support de cours pour les six paramètres :
        left, right, bottom, top, near, far */
     gl4duFrustumf(-1.0f, 1.0f, -1.0f * _wh / (GLfloat)_ww, 1.0f * _wh / (GLfloat)_ww, 1.0f, 1000.0f);
     /* binder (mettre au premier plan, "en courante" ou "en active") la
        matrice view */
     gl4duBindMatrix("view");
     /* mettre la matrice identité (celle qui ne change rien) dans la matrice courante */
     gl4duLoadIdentityf();
     /* composer (multiplication à droite) avec une translation dans les z négatifs (-5) */
     /* gl4duTranslatef(0, 0, -5); */
     gl4duLookAtf(0, 2, 5, 0, 0, 0, 0, 1, 0);
     break;

  }
}

void scene (float a){
  /* utiliser le programme GPU "_pId" */
  glUseProgram(_pId);
  /* binder (mettre au premier plan, "en courante" ou "en active") la
     matrice modview */
  gl4duBindMatrix("modview");
  /* mettre la matrice identité (celle qui ne change rien) dans la matrice courante */
  gl4duLoadIdentityf();
  gl4duPushMatrix();
  /* composer (multiplication à droite) avec une rotation d'angle a et
     d'axe (autour de l'axe) <0, 1, 0> */
  gl4duTranslatef(-2, 0, 0);
  gl4duRotatef(a, 0, 1, 0);
  /* composer (multiplication à droite) avec une translation dans les z négatifs (-5) */ 
  /* envoyer les matrice GL4D au programme GPU OpenGL (en cours) */
  gl4duSendMatrices();
  glUniform4f(_color, 0., 1., 1., 1.);
  //glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

#if 0
  /* activer l'étage de texture 0 */
  glActiveTexture(GL_TEXTURE0);
  /* binder _texId comme texture bidimensionnelle */
  glBindTexture(GL_TEXTURE_2D, _texId);
  glUniform1i(glGetUniformLocation(_pId, "tex"), 0); /* 0 correspond à GL_TEXTURE0 */
#endif 
  /* booléen pour dire d'utiliser ou pas une texture */
  glUniform1i(glGetUniformLocation(_pId, "use_texture"), 0);
  /* demander le dessin d'un objet GL4D */
  gl4dgDraw(_cube);

  gl4duPopMatrix();
  
  gl4duPushMatrix();
  /* composer (multiplication à droite) avec une rotation d'angle a et
     d'axe (autour de l'axe) <0, 1, 0> */
  gl4duTranslatef(2, 0, 0);
  gl4duRotatef(a, 1, 1, 0);
  /* composer (multiplication à droite) avec une translation dans les z négatifs (-5) */ 
  /* envoyer les matrice GL4D au programme GPU OpenGL (en cours) */
  gl4duSendMatrices();
  /* demander le dessin d'un objet GL4D */
  glUniform4f(_color, 1., 0.39, 0.28, 1.);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  glUniform1i(glGetUniformLocation(_pId, "use_texture"), 0);
  gl4dgDraw(_torus);
  gl4duPopMatrix();
  /* n'utiliser aucun programme GPU (pas nécessaire) */
  glUseProgram(0);
}

void draw(void) {
  /* une variable pour stocker un angle qui incrémente */
  static float a = 0;
  /* effacer le buffer de couleur (image) et le buffer de profondeur d'OpenGL */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  if (_multi_vues){ 
    vue (0); 
    scene(a);
    vue (1); 
    scene(a);
    vue (2); 
    scene(a);
    vue (3); 
    scene(a);
  }
  else {
    vue(4);
    scene(a);
  }
  /* augmenter l'ange a de 1 */
  a += 1;
}
/* appelée lors du exit */
void quit(void) {
  /* libérer l'identifiant de texture (et la donnée associée en GPU) */
  if(_texId) {
     glDeleteTextures(1, &_texId);
     _texId = 0;
  }
  /* nettoyer (libérer) tout objet créé avec GL4D */
  gl4duClean(GL4DU_ALL);
}

